# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mplang/protos/v1alpha1/mpir.proto
# Protobuf Python Version: 5.26.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n!mplang/protos/v1alpha1/mpir.proto\x12\x16mplang.protos.v1alpha1\"\xde\x03\n\tAttrProto\x12>\n\x04type\x18\x01 \x01(\x0e\x32*.mplang.protos.v1alpha1.AttrProto.AttrTypeR\x04type\x12\x0c\n\x01\x66\x18\x02 \x01(\x02R\x01\x66\x12\x0c\n\x01i\x18\x03 \x01(\x03R\x01i\x12\x0c\n\x01s\x18\x04 \x01(\tR\x01s\x12\x0c\n\x01\x62\x18\x05 \x01(\x08R\x01\x62\x12\x1b\n\traw_bytes\x18\x06 \x01(\x0cR\x08rawBytes\x12\x16\n\x06\x66loats\x18\x07 \x03(\x02R\x06\x66loats\x12\x12\n\x04ints\x18\x08 \x03(\x03R\x04ints\x12\x12\n\x04strs\x18\t \x03(\tR\x04strs\x12\x35\n\x04\x66unc\x18\n \x01(\x0b\x32!.mplang.protos.v1alpha1.FuncProtoR\x04\x66unc\x12\x38\n\x05graph\x18\x0b \x01(\x0b\x32\".mplang.protos.v1alpha1.GraphProtoR\x05graph\"\x8a\x01\n\x08\x41ttrType\x12\r\n\tUNDEFINED\x10\x00\x12\t\n\x05\x46LOAT\x10\x01\x12\x07\n\x03INT\x10\x02\x12\n\n\x06STRING\x10\x03\x12\x08\n\x04\x42OOL\x10\x04\x12\t\n\x05\x42YTES\x10\x05\x12\n\n\x06\x46LOATS\x10\x06\x12\x08\n\x04INTS\x10\x07\x12\x0b\n\x07STRINGS\x10\x08\x12\x0c\n\x08\x46UNCTION\x10\n\x12\t\n\x05GRAPH\x10\x0b\"\x87\x02\n\tFuncProto\x12\x12\n\x04type\x18\x01 \x01(\tR\x04type\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x12\n\x04\x62ody\x18\x03 \x01(\tR\x04\x62ody\x12\x1d\n\ndoc_string\x18\x04 \x01(\tR\tdocString\x12\x42\n\x05\x61ttrs\x18\x05 \x03(\x0b\x32,.mplang.protos.v1alpha1.FuncProto.AttrsEntryR\x05\x61ttrs\x1a[\n\nAttrsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32!.mplang.protos.v1alpha1.AttrProtoR\x05value:\x02\x38\x01\"h\n\x0fTensorTypeProto\x12\x36\n\x05\x64type\x18\x01 \x01(\x0e\x32 .mplang.protos.v1alpha1.DataTypeR\x05\x64type\x12\x1d\n\nshape_dims\x18\x02 \x03(\x03R\tshapeDims\"\xaf\x01\n\x0eTableTypeProto\x12G\n\x07\x63olumns\x18\x01 \x03(\x0b\x32-.mplang.protos.v1alpha1.TableTypeProto.ColumnR\x07\x63olumns\x1aT\n\x06\x43olumn\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x36\n\x05\x64type\x18\x02 \x01(\x0e\x32 .mplang.protos.v1alpha1.DataTypeR\x05\x64type\"\xe8\x02\n\x0bMPTypeProto\x12J\n\x0btensor_type\x18\x01 \x01(\x0b\x32\'.mplang.protos.v1alpha1.TensorTypeProtoH\x00R\ntensorType\x12G\n\ntable_type\x18\x02 \x01(\x0b\x32&.mplang.protos.v1alpha1.TableTypeProtoH\x00R\ttableType\x12\x14\n\x05pmask\x18\x03 \x01(\x03R\x05pmask\x12\x44\n\x05\x61ttrs\x18\x04 \x03(\x0b\x32..mplang.protos.v1alpha1.MPTypeProto.AttrsEntryR\x05\x61ttrs\x1a[\n\nAttrsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32!.mplang.protos.v1alpha1.AttrProtoR\x05value:\x02\x38\x01\x42\x0b\n\ttype_spec\"\xd2\x02\n\tNodeProto\x12\x17\n\x07op_type\x18\x01 \x01(\tR\x06opType\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x16\n\x06inputs\x18\x03 \x03(\tR\x06inputs\x12@\n\touts_info\x18\x05 \x03(\x0b\x32#.mplang.protos.v1alpha1.MPTypeProtoR\x08outsInfo\x12\x42\n\x05\x61ttrs\x18\x04 \x03(\x0b\x32,.mplang.protos.v1alpha1.NodeProto.AttrsEntryR\x05\x61ttrs\x12\x1d\n\ndoc_string\x18\x06 \x01(\tR\tdocString\x1a[\n\nAttrsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32!.mplang.protos.v1alpha1.AttrProtoR\x05value:\x02\x38\x01\"e\n\x0bVersionInfo\x12\x14\n\x05major\x18\x01 \x01(\x05R\x05major\x12\x14\n\x05minor\x18\x02 \x01(\x05R\x05minor\x12\x14\n\x05patch\x18\x03 \x01(\x05R\x05patch\x12\x14\n\x05\x65xtra\x18\x04 \x01(\tR\x05\x65xtra\"\xc0\x02\n\nGraphProto\x12=\n\x07version\x18\x01 \x01(\x0b\x32#.mplang.protos.v1alpha1.VersionInfoR\x07version\x12\x37\n\x05nodes\x18\x02 \x03(\x0b\x32!.mplang.protos.v1alpha1.NodeProtoR\x05nodes\x12\x18\n\x07outputs\x18\x03 \x03(\tR\x07outputs\x12\x43\n\x05\x61ttrs\x18\x04 \x03(\x0b\x32-.mplang.protos.v1alpha1.GraphProto.AttrsEntryR\x05\x61ttrs\x1a[\n\nAttrsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x37\n\x05value\x18\x02 \x01(\x0b\x32!.mplang.protos.v1alpha1.AttrProtoR\x05value:\x02\x38\x01*\xea\x01\n\x08\x44\x61taType\x12\r\n\tUNDEFINED\x10\x00\x12\x06\n\x02U8\x10\x01\x12\x06\n\x02I8\x10\x02\x12\x07\n\x03U16\x10\x03\x12\x07\n\x03I16\x10\x04\x12\x07\n\x03U32\x10\x05\x12\x07\n\x03I32\x10\x06\x12\x07\n\x03U64\x10\x07\x12\x07\n\x03I64\x10\x08\x12\x07\n\x03\x46\x31\x36\x10\t\x12\x07\n\x03\x46\x33\x32\x10\n\x12\x07\n\x03\x46\x36\x34\x10\x0b\x12\n\n\x06STRING\x10\x0c\x12\x08\n\x04\x42OOL\x10\r\x12\r\n\tCOMPLEX64\x10\x0e\x12\x0e\n\nCOMPLEX128\x10\x0f\x12\x0c\n\x08\x42\x46LOAT16\x10\x10\x12\x08\n\x04\x44\x41TE\x10\x11\x12\x08\n\x04TIME\x10\x12\x12\r\n\tTIMESTAMP\x10\x13\x12\x08\n\x04JSON\x10\x14\x42.Z,github.com/secretflow/mplang/protos/v1alpha1b\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'mplang.protos.v1alpha1.mpir_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'Z,github.com/secretflow/mplang/protos/v1alpha1'
  _globals['_FUNCPROTO_ATTRSENTRY']._loaded_options = None
  _globals['_FUNCPROTO_ATTRSENTRY']._serialized_options = b'8\001'
  _globals['_MPTYPEPROTO_ATTRSENTRY']._loaded_options = None
  _globals['_MPTYPEPROTO_ATTRSENTRY']._serialized_options = b'8\001'
  _globals['_NODEPROTO_ATTRSENTRY']._loaded_options = None
  _globals['_NODEPROTO_ATTRSENTRY']._serialized_options = b'8\001'
  _globals['_GRAPHPROTO_ATTRSENTRY']._loaded_options = None
  _globals['_GRAPHPROTO_ATTRSENTRY']._serialized_options = b'8\001'
  _globals['_DATATYPE']._serialized_start=2223
  _globals['_DATATYPE']._serialized_end=2457
  _globals['_ATTRPROTO']._serialized_start=62
  _globals['_ATTRPROTO']._serialized_end=540
  _globals['_ATTRPROTO_ATTRTYPE']._serialized_start=402
  _globals['_ATTRPROTO_ATTRTYPE']._serialized_end=540
  _globals['_FUNCPROTO']._serialized_start=543
  _globals['_FUNCPROTO']._serialized_end=806
  _globals['_FUNCPROTO_ATTRSENTRY']._serialized_start=715
  _globals['_FUNCPROTO_ATTRSENTRY']._serialized_end=806
  _globals['_TENSORTYPEPROTO']._serialized_start=808
  _globals['_TENSORTYPEPROTO']._serialized_end=912
  _globals['_TABLETYPEPROTO']._serialized_start=915
  _globals['_TABLETYPEPROTO']._serialized_end=1090
  _globals['_TABLETYPEPROTO_COLUMN']._serialized_start=1006
  _globals['_TABLETYPEPROTO_COLUMN']._serialized_end=1090
  _globals['_MPTYPEPROTO']._serialized_start=1093
  _globals['_MPTYPEPROTO']._serialized_end=1453
  _globals['_MPTYPEPROTO_ATTRSENTRY']._serialized_start=715
  _globals['_MPTYPEPROTO_ATTRSENTRY']._serialized_end=806
  _globals['_NODEPROTO']._serialized_start=1456
  _globals['_NODEPROTO']._serialized_end=1794
  _globals['_NODEPROTO_ATTRSENTRY']._serialized_start=715
  _globals['_NODEPROTO_ATTRSENTRY']._serialized_end=806
  _globals['_VERSIONINFO']._serialized_start=1796
  _globals['_VERSIONINFO']._serialized_end=1897
  _globals['_GRAPHPROTO']._serialized_start=1900
  _globals['_GRAPHPROTO']._serialized_end=2220
  _globals['_GRAPHPROTO_ATTRSENTRY']._serialized_start=715
  _globals['_GRAPHPROTO_ATTRSENTRY']._serialized_end=806
# @@protoc_insertion_point(module_scope)
